/* $Id: input.c,v 1.12 1999/06/11 23:07:41 marcus Exp $
******************************************************************************

   Input-5dtglove: all-in-one-file.
   
   Glove driver for the 5dt cyberglove. It uses the 5dt-provided driver 
   library for actual hardware access.

   Copyright (C) 2000 Andreas Beck      [becka@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <stdlib.h>

#include <ggi/internal/gii-dl.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_SYS_IO_H
#include <sys/io.h>
#endif
#ifdef HAVE_ASM_IO_H
#include <asm/io.h>
#endif

#include "fglove.h"

static char *devicename="/dev/fdglove";

struct glovestate {
	int buttonstate;
	int *axispos;
};


struct joyinfo {
	fdGlove *fdghandle;
	int numbuttons;
	int numaxis;
	double	scaleaxis;
	struct glovestate last;
	struct glovestate curr;
};

static void _GII_5dtglove_getnew(struct joyinfo *joy)
{
	int x;

	GIIDPRINT_MISC("input-5dtglove: getnew\n");
	joy->curr.buttonstate=fdGetGesture(joy->fdghandle);
	for(x=0;x<joy->numaxis;x++)
	{
		joy->curr.axispos[x]=joy->scaleaxis*fdGetSensorScaled(joy->fdghandle,x);
	}
}

static gii_event_mask GII_5dtglove_poll(gii_input *inp, void *arg)
{
	struct joyinfo *joy=inp->priv;
	gii_event_mask rc=0;
	gii_event ev;
	int x;
	
	GIIDPRINT_MISC("input-5dtglove: poll(%p);\n",inp);

	/* Get new data first. */
	_GII_5dtglove_getnew(joy);

	if (memcmp(&joy->curr,&joy->last,sizeof(joy->last))==0)
		return 0;	/* Nothing happened */

	if (joy->last.buttonstate!=joy->curr.buttonstate)
	{
		_giiEventBlank(&ev, sizeof(gii_key_event));
		ev.key.size=sizeof(gii_key_event);
		ev.key.type=joy->curr.buttonstate ?
			evKeyPress : evKeyRelease;
		rc |= joy->curr.buttonstate ?
			emKeyPress : emKeyRelease;
		GIIDPRINT_MISC("input-5dtglove: read KEY\n");
		ev.key.origin=inp->origin;
		ev.key.modifiers=0;
		ev.key.sym=ev.key.label=GIIK_VOID;
		ev.key.button=joy->curr.buttonstate ? joy->curr.buttonstate : joy->last.buttonstate;
		_giiEvQueueAdd(inp,&ev);
		joy->last.buttonstate=joy->curr.buttonstate;
	}
	for(x=0;x<joy->numaxis;x++)
		if (joy->last.axispos[x]!=joy->curr.axispos[x])
		{
			_giiEventBlank(&ev, sizeof(gii_val_event));
			ev.size=sizeof(gii_val_event);
			GIIDPRINT_MISC("input-5dtglove: read VAL\n");
			ev.val.type=evValAbsolute;
			rc |= emValAbsolute;
			ev.val.origin=inp->origin;
			ev.val.first=x;
			ev.val.count=1;
			ev.val.value[0]=joy->curr.axispos[x];
			_giiEvQueueAdd(inp,&ev);
			joy->last.axispos[x]=joy->curr.axispos[x];
		}

	joy->last=joy->curr;

	return rc;
}

static int GII_5dtglove_close(gii_input *inp)
{
	struct joyinfo *joy=inp->priv;

	fdClose(joy->fdghandle);
	free(joy->last.axispos    );
	free(joy->curr.axispos    );
	free(joy);
	
	GIIDPRINT_MISC("input-5dtglove: close %p\n",inp);
	return 0;
}

int GIIdlinit(gii_input *inp,const char *args, void *argptr)
{
	struct joyinfo *joy;
	int x;

	GIIDPRINT_MISC("input-5dtglove starting.(args=\"%s\",argptr=%p)\n",args ? args : "(nil)",argptr);

	if (NULL==(inp->priv=joy=malloc(sizeof(struct joyinfo))))
	{
		GIIDPRINT_MISC("input-5dtglove: no memory.\n");
		return -1;
	}

	if (NULL==(joy->fdghandle=fdOpen(devicename)))	/* FIXME - Make configureable */
	{
		GIIDPRINT_MISC("input-5dtglove: Cannot open glove at %s.\n",devicename);
		free(joy);
		return -1;
	}

	switch(fdGetGloveType(joy->fdghandle)) {
		case FD_GLOVENONE: 
			GIIDPRINT_MISC("input-5dtglove: no glove detected.\n");
			free(joy);return -1;
		case FD_GLOVE7:
			GIIDPRINT_MISC("input-5dtglove:  7 sensor glove detected.\n");break;
		case FD_GLOVE7W:
			GIIDPRINT_MISC("input-5dtglove:  7 sensor wireless glove detected.\n");break;
		case FD_GLOVE16:
			GIIDPRINT_MISC("input-5dtglove: 16 sensor glove detected.\n");break;
		case FD_GLOVE16W:
			GIIDPRINT_MISC("input-5dtglove: 16 sensor wireless glove detected.\n");break;
	}

	switch(fdGetGloveHand(joy->fdghandle)) {
		case FD_HAND_LEFT: GIIDPRINT_MISC("input-5dtglove: left  handed glove found.\n");break;
		case FD_HAND_RIGHT:GIIDPRINT_MISC("input-5dtglove: right handed glove found.\n");break;
	}

	joy->numbuttons=fdGetNumGestures(joy->fdghandle);
	joy->numaxis   =fdGetNumSensors(joy->fdghandle);
	joy->scaleaxis =1000;	/* FIXME */
	joy->last.buttonstate=0;
	joy->last.axispos    =malloc(joy->numaxis   *sizeof(joy->last.axispos[0]));
	joy->curr.buttonstate=0;
	joy->curr.axispos    =malloc(joy->numaxis   *sizeof(joy->curr.axispos[0]));
	for(x=0;x<joy->numaxis;x++)
		joy->last.axispos[x]=joy->curr.axispos[x]=0;

	/* We leave these on the default handlers
	 *	inp->GIIseteventmask=_GIIstdseteventmask;
	 *	inp->GIIgeteventmask=_GIIstdgeteventmask;
	 *	inp->GIIgetselectfdset=_GIIstdgetselectfd;
	 */

	/* They are already set, so we can as well use them instead of
	 * accessing the curreventmask member directly.
	 */
	inp->targetcan=emKeyPress | emKeyRelease | emValAbsolute;
	inp->GIIseteventmask(inp,emKeyPress | emKeyRelease | emValAbsolute);

	inp->maxfd=0;			/* We poll - sorry. */
	inp->flags|=GII_FLAGS_HASPOLLED;

	inp->GIIclose	 =GII_5dtglove_close;
	inp->GIIeventpoll=GII_5dtglove_poll;

	GIIDPRINT_MISC("input-5dtglove fully up\n");

	return 0;
}
